
`timescale 1ns / 1ps

module flash_ctrl(CLK,RST,INBUS,FLC_DV,FLC_OE,FLC_BG,FRFL_DV,RDY4INP,FROM_DATA,FROM_ADDR,FROM_PAGE,
                  FLC_BR,INFF_AF,WR_RDB,NEW_PAGE,PUSH,PAGE,ADDR,TO_DATA,
                  FL_TYPE,FL_TX_ODV,FL_SND_NOW,FL_CNFG_ODV,FL_CNFG_NOW_DONE,OUTBUS);


    input CLK;
    input RST;
    input [15:0] INBUS;
    input FLC_DV;
    input FLC_OE;
    input FLC_BG;
    input FRFL_DV;
    input RDY4INP;
    input [7:0] FROM_DATA;
    input [5:0] FROM_ADDR;
    input [6:0] FROM_PAGE;
    output FLC_BR;
    output INFF_AF;
    output WR_RDB;
    output NEW_PAGE;
    output PUSH;
    output [6:0] PAGE;
    output [5:0] ADDR;
    output [7:0] TO_DATA;
    output [7:0] FL_TYPE;
    output FL_TX_ODV;
    output FL_SND_NOW;
    output FL_CNFG_ODV;
    output FL_CNFG_NOW_DONE;
    output [15:0] OUTBUS;

reg FLC_BR;
reg WR_RDB;
reg NEW_PAGE,NPG,NPG1;
reg PUSH, PSH_FL,PSH_FL1;
reg [6:0] PAGE;
reg [5:0] ADDR;
reg [7:0] TO_DATA;
reg FL_SND_NOW;
reg FL_TX_ODV;
wire [15:0] OUTBUS;
wire [7:0] FL_TYPE;

`ifdef HEADER
`else
`include "header.v"
`endif

reg  VLD,VLD1;
reg  SND_NOW,SND1,SND2;
reg  CNFG_DONE,CD1,CD2;
reg  FL_CNFG_NOW_DONE;
reg  FL_CNFG_ODV;
reg  CNFG;
reg  ODV;

reg  SET_NPG;
reg  RST_NPG;
reg  READ_INFF;
wire INFF_MT;
wire INFF_FULL;
wire INFF_AMT;
wire INFF_AF;
wire INFF_WTERR;
wire INFF_RDERR;
wire [15:0] INFF_BUSS;
wire [9:0] INFF_CNT;
reg  DEC_INPG;
reg  LD_INPG;
wire [7:0] INPG_CNT;
wire LAST_PAGE;
reg  LD_BYTE;
wire [6:0] BYTE_CNT;
wire LAST_BYTE;
reg  CLR_ACCUM;
wire [9:0] ACCUM_BUSS;
reg  VALID;

reg  READ_ADFF;
wire ADFF_MT;
wire ADFF_FULL;
wire ADFF_AMT;
wire ADFF_AF;
wire ADFF_WTERR;
wire ADFF_RDERR;
wire [15:0] ADFF_BUSS;
wire [9:0] ADFF_CNT;

reg  RD_TRDS;
reg  PSH_TRDS;
wire TRDS_MT;
wire TRDS_FULL;
wire TRDS_AMT;
wire TRDS_AF;
wire TRDS_WTERR;
wire TRDS_RDERR;
wire [9:0] TRDS_BUSS;
wire [3:0] TRDS_FF_CNT;
reg  DEC_TRDS;
reg  LD_TRDS;
wire [9:0] TRDS_CNT;
wire PKT_DONE;

reg  RD_PGRD;
wire PSH_PGRD;
wire PGRD_MT;
wire PGRD_FULL;
wire PGRD_AMT;
wire PGRD_AF;
wire PGRD_WTERR;
wire PGRD_RDERR;
wire [14:0] PGRD_BUSS;
wire [5:0] PGRD_FF_CNT;
reg  DEC_RD;
reg  LD_RD;
wire [6:0] RD_CNT;
wire LAST_READ;
reg  SEL_PAGE;
reg  [6:0] PAGE_REG;

wire DATA_CMPLT;
wire NO_READS;
reg  [2:0] FLCI_ST;
reg  [3:0] F2GC_ST;
reg  VAL_CNT;
reg  VAL_BCNT;

assign FL_TYPE = 8'h0B;
assign LAST_BYTE =  (BYTE_CNT == 7'd1);
assign LAST_PAGE = (INPG_CNT == 8'd0);
assign NO_READS = (ACCUM_BUSS == 10'd0);
assign DATA_CMPLT =  (ADFF_CNT >= TRDS_CNT);
assign PKT_DONE =  (TRDS_CNT == 10'd1);
assign LAST_READ = (RD_CNT == 7'd1);
assign LD_ACCUM = LD_BYTE && !INFF_BUSS[15];
assign PSH_PGRD        = LD_ACCUM;

//////////////////////////////////////////////////////////////////
//                                                              //
// Input FIFO for data going to the Flash memory                //
//                                                              //
//////////////////////////////////////////////////////////////////

parameter IN_FIFO_p2depth = 10;
defparam IN_FIFO.p2depth = IN_FIFO_p2depth;
defparam IN_FIFO.width = 16;
defparam IN_FIFO.almostmt = 32;
defparam IN_FIFO.almostfull = (1<<IN_FIFO_p2depth) - 64;
fifo_dpXn_blk  IN_FIFO(.clk(CLK),.rst(RST),.push(FLC_DV),.pop(READ_INFF),
                       .mt(INFF_MT),.full(INFF_FULL),.amt(INFF_AMT),.af(INFF_AF),.wterr(INFF_WTERR),.rderr(INFF_RDERR),
                       .din(INBUS),.dout(INFF_BUSS),.cnt(INFF_CNT));

//////////////////////////////////////////////////////////////////
//                                                              //
// Total Reads FIFO.  Stores the total reads in one packet      //
//                                                              //
//////////////////////////////////////////////////////////////////

parameter Tot_Rds_FIFO_p2depth = 4;
defparam Tot_Rds_FIFO.p2depth = Tot_Rds_FIFO_p2depth;
defparam Tot_Rds_FIFO.width = 10;
defparam Tot_Rds_FIFO.almostmt = 2;
defparam Tot_Rds_FIFO.almostfull = (1<<Tot_Rds_FIFO_p2depth) - 2;
fifo_dpXn_dst  Tot_Rds_FIFO(.clk(CLK),.rst(RST),.push(PSH_TRDS),.pop(RD_TRDS),
                       .mt(TRDS_MT),.full(TRDS_FULL),.amt(TRDS_AMT),.af(TRDS_AF),.wterr(TRDS_WTERR),.rderr(TRDS_RDERR),
                       .din(ACCUM_BUSS),.dout(TRDS_BUSS),.cnt(TRDS_FF_CNT));

//////////////////////////////////////////////////////////////////
//                                                              //
// Page/Byte count FIFO.  Stores the total bytes for one page   //
//                                                              //
//////////////////////////////////////////////////////////////////

parameter PG_RdCnt_FIFO_p2depth = 6;
defparam PG_RdCnt_FIFO.p2depth = PG_RdCnt_FIFO_p2depth;
defparam PG_RdCnt_FIFO.width = 15;
defparam PG_RdCnt_FIFO.almostmt = 4;
defparam PG_RdCnt_FIFO.almostfull = (1<<PG_RdCnt_FIFO_p2depth) - 4;
fifo_dpXn_dst  PG_RdCnt_FIFO(.clk(CLK),.rst(RST),.push(PSH_PGRD),.pop(RD_PGRD),
                       .mt(PGRD_MT),.full(PGRD_FULL),.amt(PGRD_AMT),.af(PGRD_AF),.wterr(PGRD_WTERR),.rderr(PGRD_RDERR),
                       .din(INFF_BUSS[14:0]),.dout(PGRD_BUSS),.cnt(PGRD_FF_CNT));



defparam Read_Accum.in_width = 7;
defparam Read_Accum.out_width = 10;
accum Read_Accum(.clk(CLK),.clr(CLR_ACCUM),.ce(LD_ACCUM),.in(INFF_BUSS[6:0]),.out(ACCUM_BUSS));

defparam In_Pg_Cntr.width = 8;
cnt_dn_l In_Pg_Cntr(.clk(CLK),.dec(DEC_INPG),.r(RST),.l(LD_INPG),.din(INFF_BUSS[7:0]),.count(INPG_CNT));
defparam Byte_Cntr.width = 7;
cnt_dn_l Byte_Cntr(.clk(CLK),.dec(PSH_FL),.r(RST),.l(LD_BYTE),.din(INFF_BUSS[6:0]),.count(BYTE_CNT));
defparam Tot_Rds_Cntr.width = 10;
cnt_dn_l Tot_Rds_Cntr(.clk(CLK),.dec(DEC_TRDS),.r(RST),.l(LD_TRDS),.din(TRDS_BUSS),.count(TRDS_CNT));
defparam Rd_Cntr.width = 7;
cnt_dn_l Rd_Cntr(.clk(CLK),.dec(DEC_RD),.r(RST),.l(LD_RD),.din(PGRD_BUSS[6:0]),.count(RD_CNT));


always @(posedge CLK)
begin
   VALID <= READ_INFF;
   PSH_FL1 <= PSH_FL;
   PUSH <= PSH_FL1;
   NPG1 <= NPG & PSH_FL;
   NEW_PAGE <= NPG1;
   TO_DATA <= INFF_BUSS[7:0];
   ADDR <= INFF_BUSS[13:8];
   if(LD_BYTE)
      begin
         WR_RDB <= INFF_BUSS[15];
         PAGE <= INFF_BUSS[14:8];
      end
   else
      begin
         WR_RDB <= WR_RDB;
         PAGE <= PAGE;
      end
   if(LD_RD)
      begin
         PAGE_REG <= PGRD_BUSS[14:8];
         CNFG <= PGRD_BUSS[7];
      end
   else
      begin
         PAGE_REG <= PAGE_REG;
         CNFG <= CNFG;
      end
end



//////////////////////////////////////////////////////////////////
//                                                              //
// Address/Data FIFO for data coming from the Flash memory      //
//                                                              //
//////////////////////////////////////////////////////////////////

parameter AD_FIFO_p2depth = 10;
defparam AD_FIFO.p2depth = AD_FIFO_p2depth;
defparam AD_FIFO.width = 16;
defparam AD_FIFO.almostmt = 32;
defparam AD_FIFO.almostfull = (1<<AD_FIFO_p2depth) - 64;
fifo_dpXn_blk  AD_FIFO(.clk(CLK),.rst(RST),.push(FRFL_DV),.pop(READ_ADFF),
                       .mt(ADFF_MT),.full(ADFF_FULL),.amt(ADFF_AMT),.af(ADFF_AF),.wterr(ADFF_WTERR),.rderr(ADFF_RDERR),
                       .din({2'b00,FROM_ADDR,FROM_DATA}),.dout(ADFF_BUSS),.cnt(ADFF_CNT));


mux2to1_reg_3st Transmit_Buffer(.clk(CLK),.sel(SEL_PAGE),.ena(FLC_OE),.ain(ADFF_BUSS),.bin({1'b0,PAGE_REG,1'b0,RD_CNT}),.out(OUTBUS));


///////////////////////////////////////////////////////////////
//                                                           //
// Flash Memory Input FIFO Control State Machine             //
//                                                           //
///////////////////////////////////////////////////////////////

always @(posedge CLK or posedge RST)
begin: FLASH_MEM_INP_CTRL_FSM
   if(RST)
      FLCI_ST <= `FLCI_Idle;
   else
      case(FLCI_ST)
         `FLCI_Idle:
            begin
               if(!INFF_MT)
                  FLCI_ST <= `Rd_NumPg;
               else 
                  FLCI_ST <= `FLCI_Idle;
            end
         `Rd_NumPg:
            FLCI_ST <= `Ld_NumPg;
         `Ld_NumPg:
            begin
               if(VALID)
                  FLCI_ST <= `Rd_WPNB;
               else 
                  FLCI_ST <= `Ld_NumPg;
            end
         `Rd_WPNB:
            FLCI_ST <= `Ld_WPNB;
         `Ld_WPNB:
            begin
               if(VALID)
                  FLCI_ST <= `W4Rdy;
               else 
                  FLCI_ST <= `Ld_WPNB;
            end
         `W4Rdy:
            begin
               if(RDY4INP)
                  FLCI_ST <= `Xfer2Fl;
               else 
                  FLCI_ST <= `W4Rdy;
            end
         `Xfer2Fl:
            begin
               if(LAST_PAGE && LAST_BYTE && NO_READS)
                  FLCI_ST <= `FLCI_Idle;
               else if(LAST_PAGE && LAST_BYTE && !NO_READS)
                  FLCI_ST <= `Push_TRDS;
               else if(LAST_BYTE && !LAST_PAGE) 
                  FLCI_ST <= `Rd_WPNB;
               else if(!RDY4INP) 
                  FLCI_ST <= `W4Rdy;
               else
                  FLCI_ST <= `Xfer2Fl;
            end
         `Push_TRDS:
            FLCI_ST <= `FLCI_Idle;
         default:
            FLCI_ST <= `FLCI_Idle;
      endcase
end


always @(FLCI_ST)
begin
   case        (FLCI_ST)
      `FLCI_Idle:
         begin
            CLR_ACCUM <= 1;
         end
      default:
         begin
            CLR_ACCUM <= 0;
         end
   endcase
end
always @(FLCI_ST or INFF_MT)
begin
   case        (FLCI_ST)
      `Rd_NumPg,`Rd_WPNB,`Xfer2Fl:
         begin
            READ_INFF <= !INFF_MT;
         end
      default:
         begin
            READ_INFF <= 0;
         end
   endcase
end
always @(FLCI_ST or VALID)
begin
   case        (FLCI_ST)
      `Ld_NumPg:
         begin
            LD_INPG <= VALID;
            LD_BYTE <= 0;
         end
      `Ld_WPNB:
         begin
            LD_INPG <= 0;
            LD_BYTE <= VALID;
         end
      default:
         begin
            LD_INPG <= 0;
            LD_BYTE <= 0;
         end
   endcase
end
always @(FLCI_ST or INFF_MT)
begin
   case        (FLCI_ST)
      `Xfer2Fl:
         begin
            PSH_FL  <= !INFF_MT;
         end
      default:
         begin
            PSH_FL  <= 0;
         end
   endcase
end
always @(FLCI_ST)
begin
   case        (FLCI_ST)
      `Rd_WPNB:
         begin
            DEC_INPG <= 1;
            SET_NPG <= 1;
            RST_NPG <= 0;
         end
      `Xfer2Fl:
         begin
            DEC_INPG <= 0;
            SET_NPG <= 0;
            RST_NPG <= 1;
         end
      default:
         begin
            DEC_INPG <= 0;
            SET_NPG <= 0;
            RST_NPG <= 0;
         end
   endcase
end

always @(posedge CLK or posedge RST)
begin
   if(RST)
      NPG <= 0;
   else
      if(RST_NPG)
         NPG <= 0;
      else if(SET_NPG)
         NPG <= 1;
      else
         NPG <= NPG;
end

always @(FLCI_ST)
begin
   case        (FLCI_ST)
      `Push_TRDS:
         begin
            PSH_TRDS <= 1;
         end
      default:
         begin
            PSH_TRDS <= 0;
         end
   endcase
end


///////////////////////////////////////////////////////////////
//                                                           //
// Flash Memory Transfer to GbE State Machine                //
//                                                           //
///////////////////////////////////////////////////////////////

always @(posedge CLK or posedge RST)
begin: FLASH_2_GbE_CTRL_FSM
   if(RST)
      F2GC_ST <= `F2GC_Idle;
   else
      case(F2GC_ST)
         `F2GC_Idle:
            begin
               if(!TRDS_MT)
                  F2GC_ST <= `Rd_Counts;
               else 
                  F2GC_ST <= `F2GC_Idle;
            end
         `Rd_Counts:
            F2GC_ST <= `Ld_Counts;
         `Ld_Counts:
            begin
               if(VAL_CNT)
                  F2GC_ST <= `W4AllRds;
               else 
                  F2GC_ST <= `Ld_Counts;
            end
         `W4AllRds:
            begin
               if(DATA_CMPLT)
                  F2GC_ST <= `FL_Req_Bus;
               else 
                  F2GC_ST <= `W4AllRds;
            end
         `FL_Req_Bus:
            begin
               if(FLC_BG)
                  F2GC_ST <= `Xmit_Page;
               else 
                  F2GC_ST <= `FL_Req_Bus;
            end
         `Xmit_Page:
            F2GC_ST <= `Xmit_Reads;
         `Xmit_Reads:
            begin
               if(LAST_READ && PKT_DONE && CNFG)
                  F2GC_ST <= `Cnfg_Done;
               else if(LAST_READ && PKT_DONE && !CNFG)
                  F2GC_ST <= `FL_Snd_Pkt;
               else if(LAST_READ && !PKT_DONE)
                  F2GC_ST <= `Rd_Pg_Rd;
               else 
                  F2GC_ST <= `Xmit_Reads;
            end
         `Rd_Pg_Rd:
            F2GC_ST <= `Ld_Pg_Rd;
         `Ld_Pg_Rd:
            begin
               if(VAL_BCNT)
                  F2GC_ST <= `Xmit_Page;
               else 
                  F2GC_ST <= `Ld_Pg_Rd;
            end
         `FL_Snd_Pkt:
            F2GC_ST <= `W4Snd;
         `W4Snd:
            begin
               if(FL_SND_NOW)
                  F2GC_ST <= `FL_Rel_Bus;
               else 
                  F2GC_ST <= `W4Snd;
            end
         `Cnfg_Done:
            F2GC_ST <= `W4Cnfg;
         `W4Cnfg:
            begin
               if(FL_CNFG_NOW_DONE)
                  F2GC_ST <= `FL_Rel_Bus;
               else 
                  F2GC_ST <= `W4Cnfg;
            end
         `FL_Rel_Bus:
            begin
               if(!FLC_BG)
                  F2GC_ST <= `F2GC_Idle;
               else 
                  F2GC_ST <= `FL_Rel_Bus;
            end
         default:
            F2GC_ST <= `F2GC_Idle;
      endcase
end

always @(posedge CLK)
begin
   VAL_CNT <= RD_TRDS;
   VAL_BCNT <= RD_PGRD;
   SND1 <= SND_NOW;
   SND2 <= SND1;
   FL_SND_NOW <= SND2;
   CD1 <= CNFG_DONE;
   CD2 <= CD1;
   FL_CNFG_NOW_DONE <= CD2;
   VLD1 <= VLD;
   ODV <= VLD1 | SEL_PAGE;
end

always @(CNFG or ODV)
begin
   if(CNFG)
      begin
         FL_TX_ODV <= 0;
         FL_CNFG_ODV <= ODV;
      end
   else
      begin
         FL_TX_ODV <= ODV;
         FL_CNFG_ODV <= 0;
      end
end

always @(F2GC_ST or VAL_CNT or VAL_BCNT)
begin
   case(F2GC_ST)
      `Rd_Counts:
         begin
            RD_TRDS <= 1;
            RD_PGRD <= 1;
            LD_TRDS <= 0;
            LD_RD <= 0;
         end
      `Ld_Counts:
         begin
            RD_TRDS <= 0;
            RD_PGRD <= 0;
            LD_TRDS <= VAL_CNT;
            LD_RD <= VAL_BCNT;
         end
      `Rd_Pg_Rd:
         begin
            RD_TRDS <= 0;
            RD_PGRD <= 1;
            LD_TRDS <= 0;
            LD_RD <= 0;
         end
      `Ld_Pg_Rd:
         begin
            RD_TRDS <= 0;
            RD_PGRD <= 0;
            LD_TRDS <= 0;
            LD_RD <= VAL_BCNT;
         end
      default:
         begin
            RD_TRDS <= 0;
            RD_PGRD <= 0;
            LD_TRDS <= 0;
            LD_RD <= 0;
         end
   endcase
end


always @(F2GC_ST)
begin
   case(F2GC_ST)
      `Xmit_Page:
         begin
            SEL_PAGE <= 1;
            DEC_TRDS <= 0; 
            DEC_RD <= 0;
            READ_ADFF <= 0;
            VLD <= 0;
         end
      `Xmit_Reads:
         begin
            SEL_PAGE <= 0;
            DEC_TRDS <= 1; 
            DEC_RD <= 1;
            READ_ADFF <= 1;
            VLD <= 1;
         end
      default:
         begin
            SEL_PAGE <= 0;
            DEC_TRDS <= 0; 
            DEC_RD <= 0;
            READ_ADFF <= 0;
            VLD <= 0;
         end
   endcase
end

always @(F2GC_ST)
begin
   case(F2GC_ST)
      `FL_Req_Bus,`Xmit_Page,`Xmit_Reads,`Rd_Pg_Rd,`Ld_Pg_Rd,`FL_Snd_Pkt,`W4Snd,`Cnfg_Done,`W4Cnfg:
         begin
            FLC_BR <= 1;
         end
      default:
         begin
            FLC_BR <= 0;
         end
   endcase
end
always @(F2GC_ST)
begin
   case(F2GC_ST)
      `FL_Snd_Pkt:
         begin
            SND_NOW <= 1;
            CNFG_DONE <= 0;
         end
      `Cnfg_Done:
         begin
            SND_NOW <= 0;
            CNFG_DONE <= 1;
         end
      default:
         begin
            SND_NOW <= 0;
            CNFG_DONE <= 0;
         end
   endcase
end


endmodule
